# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
PyECVL is a Python wrapper for ECVL, the European Computer Vision Library.
"""

import os
from setuptools import setup, Extension

import pybind11
from pyecvl.version import VERSION


def to_bool(s):
    s = s.lower()
    return s != "off" and s != "false"


EXTRA_COMPILE_ARGS = ['-std=c++17', '-fvisibility=hidden']
OPENCV_LIBS = [
    "opencv_core",
    "opencv_imgcodecs",
    "opencv_imgproc",
    "opencv_photo",
]
DICOM_LIBS = [
    "dcmdata",
    "dcmimage",
    "dcmimgle",
    "dcmjpeg",
    "i2d",
    "ijg8",
    "ijg12",
    "ijg16",
    "oflog",
    "ofstd",
]
ECVL_LIBS = ["dataset", "ecvl_core"]
OTHER_LIBS = ["stdc++fs", "yaml-cpp"]
ALL_LIBS = ECVL_LIBS + OPENCV_LIBS + OTHER_LIBS
INCLUDE_DIRS = [
    "src",
    pybind11.get_include(),
    pybind11.get_include(user=True)
]
LIBRARY_DIRS = []
RUNTIME_LIBRARY_DIRS = []
EDDL_DIR = os.getenv("EDDL_DIR")
if EDDL_DIR:
    INCLUDE_DIRS.append(os.path.join(EDDL_DIR, "include"))
    LIBRARY_DIRS.append(os.path.join(EDDL_DIR, "lib"))
    RUNTIME_LIBRARY_DIRS.append(os.path.join(EDDL_DIR, "lib"))
ECVL_DIR = os.getenv("ECVL_DIR")
if ECVL_DIR:
    INCLUDE_DIRS.append(os.path.join(ECVL_DIR, "include"))
    LIBRARY_DIRS.append(os.path.join(ECVL_DIR, "lib"))
    RUNTIME_LIBRARY_DIRS.append(os.path.join(ECVL_DIR, "lib"))

# optional modules, on by default. Set env var to "OFF" or "FALSE" to disable
ECVL_WITH_DICOM = to_bool(os.getenv("ECVL_WITH_DICOM", "ON"))
if ECVL_WITH_DICOM:
    EXTRA_COMPILE_ARGS.append('-DECVL_WITH_DICOM')
    ALL_LIBS.extend(DICOM_LIBS)
ECVL_WITH_OPENSLIDE = to_bool(os.getenv("ECVL_WITH_OPENSLIDE", "ON"))
if ECVL_WITH_OPENSLIDE:
    EXTRA_COMPILE_ARGS.append('-DECVL_WITH_OPENSLIDE')
    ALL_LIBS.append("openslide")
ECVL_EDDL = to_bool(os.getenv("ECVL_EDDL", "ON"))
if ECVL_EDDL:
    EXTRA_COMPILE_ARGS.append('-DECVL_EDDL')
    ALL_LIBS.extend(["ecvl_eddl", "eddl"])


ext = Extension(
    "pyecvl._core",
    sources=["src/_core.cpp"],
    include_dirs=INCLUDE_DIRS,
    library_dirs=LIBRARY_DIRS,
    runtime_library_dirs=RUNTIME_LIBRARY_DIRS,
    libraries=ALL_LIBS,
    extra_compile_args=EXTRA_COMPILE_ARGS,
)


setup(
    name="pyecvl",
    version=VERSION,
    url="https://github.com/deephealthproject/pyecvl",
    description="Python wrapper for ECVL",
    long_description=__doc__,
    author="Simone Leo",
    author_email="<simone.leo@crs4.it>",
    license="MIT",
    platforms=["Linux"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
    ],
    packages=["pyecvl"],
    ext_modules=[ext],
    install_requires=["setuptools", "pybind11<2.6", "numpy"],
    zip_safe=False,
)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Performs image processing on an input image.
"""

import argparse
import os
import sys

import numpy as np
import pyecvl.ecvl as ecvl


def main(args):
    head, ext = os.path.splitext(os.path.basename(args.in_fn))
    tmp = ecvl.Image.empty()
    img = ecvl.ImRead(args.in_fn)

    print("Resizing image by new width, height")
    ecvl.ResizeDim(img, tmp, [225, 300], ecvl.InterpolationType.nearest)
    ecvl.ImWrite("%s_resized%s" % (head, ext), tmp)

    print("Resizing image by scale factor")
    ecvl.ResizeScale(img, tmp, [2, 2], ecvl.InterpolationType.cubic)
    ecvl.ImWrite("%s_resized_scale%s" % (head, ext), tmp)

    print("Flipping image")
    ecvl.Flip2D(img, tmp)
    ecvl.ImWrite("%s_flipped%s" % (head, ext), tmp)

    print("Mirroring image")
    ecvl.Mirror2D(img, tmp)
    ecvl.ImWrite("%s_mirrored%s" % (head, ext), tmp)

    print("Rotating image")
    ecvl.Rotate2D(img, tmp, 60)
    ecvl.ImWrite("%s_rotated%s" % (head, ext), tmp)

    print("Rotating image with scaling and resizing")
    ecvl.RotateFullImage2D(img, tmp, 60, 1.5)
    ecvl.ImWrite("%s_rotated_full%s" % (head, ext), tmp)

    print("Changing color space to gray")
    ecvl.ChangeColorSpace(img, tmp, ecvl.ColorType.GRAY)
    ecvl.ImWrite("%s_gray%s" % (head, ext), tmp)

    print("Calculating Otsu threshold")  # image must be gray
    thr = ecvl.OtsuThreshold(tmp)
    print("  %d" % thr)

    print("Applying fixed threshold")
    ecvl.Threshold(tmp, tmp, thr, 255)
    ecvl.ImWrite("%s_threshold%s" % (head, ext), tmp)

    print("Labeling connected components")
    labels = ecvl.Image.empty()
    ecvl.ConnectedComponentsLabeling(tmp, labels)
    ecvl.ImWrite("%s_labels%s" % (head, ext), labels)

    print("Finding contours")
    contours = ecvl.FindContours(tmp)
    print("  [%r ... %r]" % (contours[0], contours[-1]))

    print("Applying filter")
    # kernel must be float64, "xyc" and with one color channel
    kernel = ecvl.Image(
        [3, 3, 1], ecvl.DataType.float64, "xyc", ecvl.ColorType.GRAY
    )
    a = np.array(kernel, copy=False)
    a.fill(0.11)
    ecvl.Filter2D(img, tmp, kernel)
    ecvl.ImWrite("%s_filter%s" % (head, ext), tmp)

    print("Applying separable filter")
    ecvl.SeparableFilter2D(img, tmp, [1, 2, 1], [1, 0, -1])
    ecvl.ImWrite("%s_separable_filter%s" % (head, ext), tmp)

    print("Blurring with gaussian kernel")
    ecvl.GaussianBlur(img, tmp, 5, 5, 0.0)
    ecvl.ImWrite("%s_gaussian_blur%s" % (head, ext), tmp)

    print("Adding Laplace noise")
    stddev = 255 * 0.05
    ecvl.AdditiveLaplaceNoise(img, tmp, stddev)
    ecvl.ImWrite("%s_laplace_noise%s" % (head, ext), tmp)

    print("Adjusting contrast")
    gamma = 3
    ecvl.GammaContrast(img, tmp, gamma)
    ecvl.ImWrite("%s_gamma_contrast%s" % (head, ext), tmp)

    print("Setting rectangles to zero (coarse dropout)")
    prob, drop_size, per_channel = 0.5, 0.1, True
    ecvl.CoarseDropout(img, tmp, prob, drop_size, per_channel)
    ecvl.ImWrite("%s_coarse_dropout%s" % (head, ext), tmp)

    print("Horizontal concatenation")
    images = [ecvl.ImRead(args.in_fn), ecvl.ImRead(args.in_fn_2)]
    new_h = int(images[0].dims_[1] / images[1].dims_[1] * images[1].dims_[0])
    ecvl.ResizeDim(images[1], images[1], [new_h, images[0].dims_[1]])
    ecvl.HConcat(images, tmp)
    ecvl.ImWrite("hconcat.png", tmp)

    print("Vertical concatenation")
    images = [ecvl.ImRead(args.in_fn), ecvl.ImRead(args.in_fn_2)]
    new_w = int(images[0].dims_[0] / images[1].dims_[0] * images[1].dims_[1])
    ecvl.ResizeDim(images[1], images[1], [images[0].dims_[0], new_w])
    ecvl.VConcat(images, tmp)
    ecvl.ImWrite("vconcat.png", tmp)

    print("Stack along depth dimension")
    images = [ecvl.ImRead(args.in_fn), ecvl.ImRead(args.in_fn_2)]
    ecvl.ResizeDim(images[1], images[1],
                   [images[0].dims_[0], images[0].dims_[1]])
    ecvl.Stack(images, tmp)

    print("Morphology")
    kernel = ecvl.Image(
        [3, 3, 1], ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR
    )
    a = np.array(kernel, copy=False)
    a.fill(0.11)
    ecvl.Morphology(img, tmp, ecvl.MorphType.MORPH_BLACKHAT, kernel)
    ecvl.ImWrite("%s_morphology%s" % (head, ext), tmp)

    print("Inpaint")
    ecvl.CopyImage(img, tmp)
    a = np.array(tmp, copy=False)
    a[150:200, 100:150, :] = 0
    ecvl.ImWrite("%s_inpaint_before%s" % (head, ext), tmp)
    dims = tmp.dims_[:2] + [1]
    mask = ecvl.Image(dims, ecvl.DataType.uint8, "xyc", ecvl.ColorType.BGR)
    b = np.array(mask, copy=False)
    b.fill(0)
    b[150:200, 100:150, :] = 1
    ecvl.Inpaint(tmp, tmp, mask, 3)
    ecvl.ImWrite("%s_inpaint_after%s" % (head, ext), tmp)

    print("Transpose")
    ecvl.Transpose(img, tmp)
    ecvl.ImWrite("%s_transpose%s" % (head, ext), tmp)

    print("GridDistortion")
    ecvl.GridDistortion(img, tmp)
    ecvl.ImWrite("%s_grid_distortion%s" % (head, ext), tmp)

    print("ElasticTransform")
    ecvl.ElasticTransform(img, tmp)
    ecvl.ImWrite("%s_elastic_transform%s" % (head, ext), tmp)

    print("OpticalDistortion")
    ecvl.OpticalDistortion(img, tmp)
    ecvl.ImWrite("%s_optical_distortion%s" % (head, ext), tmp)

    print("Salt")
    ecvl.Salt(img, tmp, 0.1)
    ecvl.ImWrite("%s_salt%s" % (head, ext), tmp)

    print("Pepper")
    ecvl.Pepper(img, tmp, 0.1)
    ecvl.ImWrite("%s_pepper%s" % (head, ext), tmp)

    print("SaltAndPepper")
    ecvl.SaltAndPepper(img, tmp, 0.1)
    ecvl.ImWrite("%s_salt_and_pepper%s" % (head, ext), tmp)

    print("Normalize")
    ecvl.Normalize(img, tmp, 20, 1.5)
    ecvl.ImWrite("%s_normalize%s" % (head, ext), tmp)

    print("CenterCrop")
    ecvl.CenterCrop(img, tmp, [200, 300])
    ecvl.ImWrite("%s_center_crop%s" % (head, ext), tmp)

    print("ScaleTo")
    ecvl.ScaleTo(img, tmp, 100, 155)
    ecvl.ImWrite("%s_scale_to%s" % (head, ext), tmp)

    print("Applying multiple thresholds")
    ecvl.ChangeColorSpace(img, tmp, ecvl.ColorType.GRAY)
    thresholds = ecvl.OtsuMultiThreshold(tmp, 2)
    print("  otsu thresholds: %r" % (thresholds,))
    ecvl.MultiThreshold(tmp, tmp, thresholds)
    ecvl.ImWrite("%s_multi_threshold%s" % (head, ext), tmp)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("in_fn", metavar="INPUT_PATH")
    parser.add_argument("in_fn_2", metavar="INPUT_PATH_2")
    main(parser.parse_args(sys.argv[1:]))



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .image import ColorType
from . import _core
_ecvl = _core.ecvl


if not _ecvl.ECVL_EDDL:
    raise ImportError("extension module not compiled with EDDL support")


__all__ = [
    "DatasetAugmentations",
    "DLDataset",
    "ImageToTensor",
    "TensorToImage",
    "TensorToView",
]


class DatasetAugmentations(_ecvl.DatasetAugmentations):
    """\
    Represents the augmentations which will be applied to each split.
    """
    def __init__(self, augs):
        """\
        :param augs: three-element list containing the augmentations to be \
          applied to the training, validation and test splits respectively. \
          If no augmentations need to be defined for a split, set the \
          corresponding list element to None.
        """
        _ecvl.DatasetAugmentations.__init__(self, augs)

    def Apply(self, st, img, gt=None):
        """\
        Apply augmentations for the specified split to an image.

        :param st: a SplitType specifying which set of augmentations
          should be applied
        :param img: image to which augmentations should be applied
        :param gt: ground truth image to which augmentations should be applied
        :return: None
        """
        if gt is None:
            return _ecvl.DatasetAugmentations.Apply(self, st, img)
        return _ecvl.DatasetAugmentations.Apply(self, st, img, gt)


class DLDataset(_ecvl.DLDataset):
    r"""\
    DeepHealth deep learning dataset.

    :var batch_size\_: size of each dataset mini batch
    :var n_channels\_: number of image channels
    :var n_channels_gt\_: number of ground truth image channels
    :var current_split\_: = Current SplitType from which images are loaded
    :var resize_dims\_: dimensions ``[H, W]`` to which images must be resized
    :var current_batch\_: number of batches already loaded for each split
    :var ctype\_: ColorType of the dataset images
    :var ctype_gt\_: ColorType of the dataset ground truth images
    :var augs\_: DatasetAugmentations to be applied to the dataset images
      (and ground truth if existing) for each split
    """
    def __init__(self, filename, batch_size, augs=None, ctype=ColorType.BGR,
                 ctype_gt=ColorType.GRAY, verify=False):
        """\
        :param filename: path to the dataset file
        :param batch_size: size of each dataset mini batch
        :param augs: a DatasetAugmentations object specifying the training,
          validation and test augmentations to be applied to the dataset images
          (and ground truth if existing) for each split. Set to None if no
          augmentation is required
        :param ctype: ColorType of the dataset images
        :param ctype_gt: ColorType of the dataset ground truth images
        :param verify: if True, verify image existence
        """
        if augs is None:
            augs = _ecvl.DatasetAugmentations()
        _ecvl.DLDataset.__init__(
            self, filename, batch_size, augs, ctype, ctype_gt
        )

    def GetSplit(self, split=None):
        """\
        Get the image indexes for a split.

        :param split: a SplitType; if None, will default to the current split
        :return: list of image indexes for the split
        """
        if split is None:
            return _ecvl.DLDataset.GetSplit(self)
        return _ecvl.DLDataset.GetSplit(self, split)

    def ResetCurrentBatch(self):
        """\
        Reset the batch counter of the current split.
        :return: None
        """
        return _ecvl.DLDataset.ResetCurrentBatch(self)

    def ResetAllBatches(self):
        """\
        Reset the batch counter of each split.
        :return: None
        """
        return _ecvl.DLDataset.ResetAllBatches(self)

    def SetSplit(self, split):
        """\
        Set the current Split.

        :param split: a SplitType
        :return: None
        """
        return _ecvl.DLDataset.SetSplit(self, split)

    def LoadBatch(self, images, labels=None):
        """\
        Load a batch into the images and labels tensors.

        :param images: a Tensor to store the batch of images
        :param labels: a Tensor to store the batch of labels
        """
        if labels is None:
            return _ecvl.DLDataset.LoadBatch(self, images)
        return _ecvl.DLDataset.LoadBatch(self, images, labels)


# FIXME: the version with offset does not make sense, since we are returning
# the output tensor
def ImageToTensor(img, t=None, offset=None):
    """\
    Convert an ECVL Image to an EDDL Tensor.

    The input image must have 3 dimensions "xy[czo]" (in any order). The output
    tensor's dimensions will be C x H x W, where:

    * C = channels
    * H = height
    * W = width

    If ``t`` and ``offset`` are None, a new tensor is created with the above
    shape.

    If ``t`` and ``offset`` are not None, the data are inserted into the
    existing ``t`` tensor at the specified offset. This allows to insert more
    than one image into a tensor, specifying how many images are already
    stored in it.

    :param img: input image
    :param t: output tensor
    :param offset: how many images are already stored in the tensor
    :return: output tensor
    """
    if t is None and offset is None:
        return _ecvl.ImageToTensor(img)
    else:
        _ecvl.ImageToTensor(img, t, offset)
        return t


def TensorToImage(t):
    """\
    Convert an EDDL Tensor into an ECVL Image.

    Tensor dimensions must be C x H x W or N x C x H x W, where:

    * N = batch size
    * C = channels
    * H = height
    * W = width

    The output image will be "xyo" with DataType.float32 and ColorType.none.

    :param t: input tensor.
    :return: output image
    """
    return _ecvl.TensorToImage(t)


def TensorToView(t):
    return _ecvl.TensorToView(t)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .datatype import DataType
from . import _core
_ecvl = _core.ecvl


__all__ = [
    "ThresholdingType",
    "InterpolationType",
    "MorphType",
    "InpaintType",
    "BorderType",
    "ResizeDim",
    "ResizeScale",
    "Flip2D",
    "Mirror2D",
    "Rotate2D",
    "RotateFullImage2D",
    "ChangeColorSpace",
    "Threshold",
    "MultiThreshold",
    "OtsuThreshold",
    "OtsuMultiThreshold",
    "Filter2D",
    "SeparableFilter2D",
    "GaussianBlur",
    "GaussianBlur2",
    "AdditiveLaplaceNoise",
    "AdditivePoissonNoise",
    "GammaContrast",
    "CoarseDropout",
    "IntegralImage",
    "NonMaximaSuppression",
    "GetMaxN",
    "ConnectedComponentsLabeling",
    "FindContours",
    "Stack",
    "HConcat",
    "VConcat",
    "Morphology",
    "Inpaint",
    "MeanStdDev",
    "Transpose",
    "GridDistortion",
    "ElasticTransform",
    "OpticalDistortion",
    "Salt",
    "Pepper",
    "SaltAndPepper",
    "SliceTimingCorrection",
    "Moments",
    "CentralMoments",
    "DrawEllipse",
    "DropColorChannel",
    "Normalize",
    "CenterCrop",
    "ScaleTo",
]


class ThresholdingType(_ecvl.ThresholdingType):
    """\
    Enum class representing the possible thresholding types.
    """
    BINARY = _ecvl.ThresholdingType.BINARY
    BINARY_INV = _ecvl.ThresholdingType.BINARY_INV


class InterpolationType(_ecvl.InterpolationType):
    """\
    Enum class representing the possible interpolation types.
    """
    nearest = _ecvl.InterpolationType.nearest
    linear = _ecvl.InterpolationType.linear
    area = _ecvl.InterpolationType.area
    cubic = _ecvl.InterpolationType.cubic
    lanczos4 = _ecvl.InterpolationType.lanczos4


class MorphType(_ecvl.MorphType):
    """\
    Enum class representing the possible morph types.
    """
    MORPH_ERODE = _ecvl.MorphType.MORPH_ERODE
    MORPH_DILATE = _ecvl.MorphType.MORPH_DILATE
    MORPH_OPEN = _ecvl.MorphType.MORPH_OPEN
    MORPH_CLOSE = _ecvl.MorphType.MORPH_CLOSE
    MORPH_GRADIENT = _ecvl.MorphType.MORPH_GRADIENT
    MORPH_TOPHAT = _ecvl.MorphType.MORPH_TOPHAT
    MORPH_BLACKHAT = _ecvl.MorphType.MORPH_BLACKHAT
    MORPH_HITMISS = _ecvl.MorphType.MORPH_HITMISS


class InpaintType(_ecvl.InpaintType):
    """\
    Enum class representing the possible inpaint types.
    """
    INPAINT_NS = _ecvl.InpaintType.INPAINT_NS
    INPAINT_TELEA = _ecvl.InpaintType.INPAINT_TELEA


class BorderType(_ecvl.BorderType):
    """\
    Enum class representing the possible border types.
    """
    BORDER_CONSTANT = _ecvl.BorderType.BORDER_CONSTANT
    BORDER_REPLICATE = _ecvl.BorderType.BORDER_REPLICATE
    BORDER_REFLECT = _ecvl.BorderType.BORDER_REFLECT
    BORDER_WRAP = _ecvl.BorderType.BORDER_WRAP
    BORDER_REFLECT_101 = _ecvl.BorderType.BORDER_REFLECT_101
    BORDER_TRANSPARENT = _ecvl.BorderType.BORDER_TRANSPARENT


def ResizeDim(src, dst, newdims, interp=InterpolationType.linear):
    """\
    Resize an image to the specified dimensions.

    Resizes ``src`` and outputs the result in ``dst``.

    :param src: source image
    :param dst: destination image
    :param newdims: list of integers specifying the new size of each
      dimension. The lenght of the list must match the ``src`` image
      dimensions, excluding the color channel
    :param interp: InterpolationType to be used
    :return: None
    """
    return _ecvl.ResizeDim(src, dst, newdims, interp)


def ResizeScale(src, dst, scales, interp=InterpolationType.linear):
    """\
    Resize an image by scaling the dimensions by a given scale factor.

    Resizes ``src`` and outputs the result in ``dst``.

    :param src: source image
    :param dst: destination image
    :param scales: list of floats that specifies the scale to apply to each
      dimension. The length of the list must match the ``src`` image
      dimensions, excluding the color channel
    :param interp: InterpolationType to be used
    :return: None
    """
    return _ecvl.ResizeScale(src, dst, scales, interp)


def Flip2D(src, dst):
    """\
    Flip an image vertically.

    :param src: source image
    :param dst: destination image
    :return: None
    """
    return _ecvl.Flip2D(src, dst)


def Mirror2D(src, dst):
    """\
    Flip an image horizontally.

    :param src: source image
    :param dst: destination image
    :return: None
    """
    return _ecvl.Mirror2D(src, dst)


def Rotate2D(src, dst, angle, center=None, scale=1.0,
             interp=InterpolationType.linear):
    """\
    Rotate an image without changing its dimensions.

    Rotates an image clockwise by a given angle (in degrees), with respect to
    a given center. The values of unknown pixels in the output image are set
    to 0. The output image is guaranteed to have the same dimensions as the
    input one. An optional scale parameter can be provided: this won't change
    the output image size, but the image will be scaled during rotation.

    :param src: source image
    :param dst: destination image
    :param angle: the rotation angle in degrees
    :param center: a list of floats representing the coordinates of the
      rotation center. If None, the center of the image is used
    :param scale: scaling factor
    :param interp: InterpolationType to be used
    :return: None
    """
    if center is None:
        center = []
    return _ecvl.Rotate2D(src, dst, angle, center, scale, interp)


def RotateFullImage2D(src, dst, angle, scale=1.0,
                      interp=InterpolationType.linear):
    """\
    Rotate an image resizing the output to fit all the pixels.

    Rotates an image clockwise by a given angle (in degrees). The values of
    unknown pixels in the output image are set to 0. The output Image is
    guaranteed to contain all the pixels of the rotated image. Thus, its
    dimensions can be different from those of the input one. An optional scale
    parameter can be provided: if set, the image will also be scaled.

    :param src: source image
    :param dst: destination image
    :param angle: the rotation angle in degrees
    :param scale: scaling factor
    :param interp: InterpolationType to be used
    :return: None
    """
    return _ecvl.RotateFullImage2D(src, dst, angle, scale, interp)


def ChangeColorSpace(src, dst, new_type):
    """\
    Copy the ``src`` image into the ``dst`` image changing the color space.

    Source and destination can be the same image.

    :param src: source image
    :param dst: destination image
    :param new_type: a ColorType specifying the new color space
    :return: None
    """
    return _ecvl.ChangeColorSpace(src, dst, new_type)


def Threshold(src, dst, thresh, maxval, thresh_type=ThresholdingType.BINARY):
    """\
    Apply a fixed threshold to an image.

    This function can be used to get a binary image out of a grayscale
    (ColorType.GRAY) image or to remove noise, filtering out pixels with too
    small or too large values. Pixels up to the ``thresh`` value will be set
    to 0, others will be set to ``maxval`` if ``thresh_type`` is
    ThresholdingType.BINARY. The opposite will happen if ``thresh_type`` is
    set to ThresholdingType.BINARY_INV.

    :param src: source image
    :param dst: destination image
    :param thresh: threshold value
    :param maxval: maximum values in the thresholded image
    :param thresh_type: ThresholdingType to be applied
    :return: None
    """
    return _ecvl.Threshold(src, dst, thresh, maxval, thresh_type)


def MultiThreshold(src, dst, thresholds, minval=0, maxval=255):
    """\
    Apply multiple thresholds to an image.

    The resulting image is quantized based on the provided ``thresholds``
    values. Output values will range uniformly from ``minval`` to ``maxval``.

    :param src: source image
    :param dst: destination image
    :param thresh: threshold values
    :param maxval: minimum value in the output image
    :param maxval: maximum value in the output image
    :return: None
    """
    return _ecvl.MultiThreshold(src, dst, thresholds, minval, maxval)


def OtsuThreshold(src):
    """\
    Calculate the Otsu thresholding value.

    The image must be ColorType.GRAY.

    :param src: source image
    :return: Otsu threshold value
    """
    return _ecvl.OtsuThreshold(src)


def OtsuMultiThreshold(src, n_thresholds=2):
    """\
    Calculate the Otsu thresholding values.

    The image must be ColorType.GRAY. The number of thresholds to be found is
    defined by the n_thresholds parameter (default is 2).

    :param src: source image
    :return: Otsu threshold values (list of integers)
    """
    return _ecvl.OtsuMultiThreshold(src, n_thresholds)


def Filter2D(src, dst, ker, type=DataType.none):
    """\
    Convolve an image with a kernel.

    :param src: source image
    :param dst: destination image
    :param ker: convolution kernel
    :param type: destination DataType. If set to DataType.none, the DataType
      of ``src`` is used
    :return: None
    """
    return _ecvl.Filter2D(src, dst, ker, type)


def SeparableFilter2D(src, dst, kerX, kerY, type=DataType.none):
    """\
    Convolve an image with a couple of 1-dimensional kernels.

    :param src: source image
    :param dst: destination image
    :param kerX: convolution kernel for the X axis.
    :param kerY: convolution kernel for the Y axis.
    :param type: destination DataType. If set to DataType.none, the DataType
      of ``src`` is used
    :return: None
    """
    return _ecvl.SeparableFilter2D(src, dst, kerX, kerY, type)


def GaussianBlur(src, dst, sizeX, sizeY, sigmaX, sigmaY=0):
    """\
    Blurs an image using a Gaussian kernel.

    :param src: source image
    :param dst: destination image
    :param sizeX: horizontal size of the kernel. Must be positive and odd
    :param sizeY: vertical size of the kernel. Must be positive and odd
    :param sigmaX: Gaussian kernel standard deviation in the X direction.
    :param sigmaY: Gaussian kernel standard deviation in the Y direction. If
      zero, sigmaX is used. If both are zero, they are calculated from sizeX
      and sizeY.
    :return: None
    """
    return _ecvl.GaussianBlur(src, dst, sizeX, sizeY, sigmaX, sigmaY)


def GaussianBlur2(src, dst, sigma):
    return _ecvl.GaussianBlur(src, dst, sigma)


def AdditiveLaplaceNoise(src, dst, std_dev):
    """\
    Adds Laplace distributed noise to an image.

    :param src: source image
    :param dst: destination image
    :param std_dev: standard deviation of the noise-generating distribution.
      Suggested values are around 255 * 0.05 for uint8 images
    :return: None
    """
    return _ecvl.AdditiveLaplaceNoise(src, dst, std_dev)


def AdditivePoissonNoise(src, dst, lambda_):
    r"""\
    Adds Poisson distributed noise to an image.

    :param src: source image
    :param dst: destination image
    :param lambda\_: lambda parameter of the Poisson distribution
    :return: None
    """
    return _ecvl.AdditivePoissonNoise(src, dst, lambda_)


def GammaContrast(src, dst, gamma):
    """\
    Adjust contrast by scaling each pixel value X to 255 * ((X/255) ** gamma).

    :param src: source image
    :param dst: destination image
    :param gamma: exponent for the contrast adjustment
    :return: None
    """
    return _ecvl.GammaContrast(src, dst, gamma)


def CoarseDropout(src, dst, p, drop_size, per_channel):
    """\
    Set rectangular areas within an image to zero.

    :param src: source image
    :param dst: destination image
    :param p: probability of any rectangle being set to zero
    :param drop_size: size of rectangles in percentage of the input image
    :param per_channel: whether to use the same value for all channels
    :return: None
    """
    return _ecvl.CoarseDropout(src, dst, p, drop_size, per_channel)


def IntegralImage(src, dst, dst_type=DataType.float64):
    """\
    Calculate the integral image of the source image.

    The ``src`` image must be ColorType.GRAY, "xyc" and DataType.uint8.

    :param src: source image
    :param dst: destination image
    :param dst_type: DataType of the destination image
    :return: None
    """
    return _ecvl.IntegralImage(src, dst, dst_type)


def NonMaximaSuppression(src, dst):
    """\
    Calculate the non-maxima suppression of the source image.

    The ``src`` image must be ColorType.GRAY, "xyc" and DataType.int32.

    :param src: source image
    :param dst: destination image
    :return: None
    """
    return _ecvl.NonMaximaSuppression(src, dst)


def GetMaxN(src, n):
    """\
    Get the indices of the ``n`` maximum values of an image.

    The ``src`` image must be ColorType.GRAY, "xyc" and DataType.int32.

    :param src: source image
    :param n: how many values to return
    :return: list of pairs corresponding to the coordinates of the max values
    """
    return _ecvl.GetMaxN(src, n)


def ConnectedComponentsLabeling(src, dst):
    """\
    Label connected components in the input image.

    The ``src`` image must be "xyc", only one color channel and DataType.uint8.

    :param src: source image
    :param dst: destination image
    :return: None
    """
    return _ecvl.ConnectedComponentsLabeling(src, dst)


def FindContours(src):
    """\
    Find contours in the input image.

    The ``src`` image must be "xyc", only one color channel and DataType.uint8.

    :param src: source image
    :return: image contours
    """
    return _ecvl.FindContours(src)


def Stack(src, dst):
    """\
    Stack a sequence of images along a new depth dimension.

    Images must be "xyc" and their dimensions must match.

    :param src: list of source images
    :param dst: destination image
    :return: None
    """
    return _ecvl.Stack(src, dst)


def HConcat(src, dst):
    """\
    Concatenate images horizontally.

    Images must be "xyc" and have the same number of rows.

    :param src: list of source images
    :param dst: destination image
    :return: None
    """
    return _ecvl.HConcat(src, dst)


def VConcat(src, dst):
    """\
    Concatenate images vertically.

    Images must be "xyc" and have the same number of columns.

    :param src: list of source images
    :param dst: destination image
    :return: None
    """
    return _ecvl.VConcat(src, dst)


def Morphology(src, dst, op, kernel, anchor=None, iterations=1,
               border_type=BorderType.BORDER_CONSTANT, border_value=0):
    """\
    Perform morphological transformations based on erosion and dilation.

    :param src: input image
    :param dst: output image
    :param op: a MorphType
    :param kernel: structuring element
    :param anchor: anchor position within the kernel. A negative value means
      that the anchor is at the center of the kernel
    :param iterations: number of times erosion and dilation are applied
    :param borderType: pixel extrapolation method, see BorderType.
      BorderType.BORDER_WRAP is not supported
    :param borderValue: border value in case of a constant border
    :return: None
    """
    if anchor is None:
        anchor = [-1, -1]
    return _ecvl.Morphology(src, dst, op, kernel, anchor, iterations,
                            border_type, border_value)


def Inpaint(src, dst, inpaintMask, inpaintRadius,
            flag=InpaintType.INPAINT_TELEA):
    """\
    Restore a region in an image using the region's neighborhood.

    :param src: input image
    :param dst: output image
    :param inpaintMask: an Image with 1 channel and DataType.uint8. Non-zero
      pixels indicate the area that needs to be inpainted.
    :param inpaintRadius: radius of a circular neighborhood of each point
      inpainted that is considered by the algorithm.
    :param flag: inpainting method (an InpaintType)
    :return: None
    """
    return _ecvl.Inpaint(src, dst, inpaintMask, inpaintRadius, flag)


def MeanStdDev(src):
    """\
    Calculate the mean and the standard deviation of an image.

    :param src: input image
    :return: a (mean, stddev) tuple
    """
    return _ecvl.MeanStdDev(src)


def Transpose(src, dst):
    """\
    Swap rows and columns of an image.

    :param src: input image
    :param dst: output image
    :return: None
    """
    return _ecvl.Transpose(src, dst)


def GridDistortion(src, dst, num_steps=5, distort_limit=None,
                   interp=InterpolationType.linear,
                   border_type=BorderType.BORDER_REFLECT_101, border_value=0,
                   seed=None):
    """\
    Divide the image into a cell grid and randomly stretch or reduce each cell.

    :param src: input image
    :param dst: output image
    :param num_steps: grid cell count on each side
    :param distort_limit: distortion steps range
    :param interp: InterpolationType to be used
    :param border_type: pixel extrapolation method, see BorderType
    :param border_value: padding value if border_type is
      BorderType.BORDER_CONSTANT
    :param seed: seed for the random number generator
    :return: None
    """
    if distort_limit is None:
        distort_limit = [-0.3, 0.3]
    if seed is None:
        return _ecvl.GridDistortion(src, dst, num_steps, distort_limit,
                                    interp, border_type, border_value)
    return _ecvl.GridDistortion(src, dst, num_steps, distort_limit,
                                interp, border_type, border_value, seed)


def ElasticTransform(src, dst, alpha=34, sigma=4,
                     interp=InterpolationType.linear,
                     border_type=BorderType.BORDER_REFLECT_101,
                     border_value=0, seed=None):
    """\
    Elastic deformation of input image.

    :param src: input image
    :param dst: output image
    :param alpha: scaling factor that controls the intensity of the deformation
    :param sigma: Gaussian kernel standard deviation
    :param interp: InterpolationType to be used. If ``src`` is DataType.int8
      or DataType.int32, InterpolationType.nearest is used
    :param border_type: pixel extrapolation method, see BorderType
    :param border_value: padding value if border_type is
      BorderType.BORDER_CONSTANT
    :param seed: seed for the random number generator
    :return: None
    """
    if seed is None:
        return _ecvl.ElasticTransform(src, dst, alpha, sigma, interp,
                                      border_type, border_value)
    return _ecvl.ElasticTransform(src, dst, alpha, sigma, interp, border_type,
                                  border_value, seed)


def OpticalDistortion(src, dst, distort_limit=None, shift_limit=None,
                      interp=InterpolationType.linear,
                      border_type=BorderType.BORDER_REFLECT_101,
                      border_value=0, seed=None):
    """\
    Barrel / pincushion distortion.

    :param src: input image
    :param dst: output image
    :param distort_limit: distortion intensity range
    :param shift_limit: image shifting range
    :param interp: InterpolationType to be used
    :param border_type: pixel extrapolation method, see BorderType
    :param border_value: padding value if border_type is
      BorderType.BORDER_CONSTANT
    :param seed: seed for the random number generator
    :return: None
    """
    if distort_limit is None:
        distort_limit = [-0.3, 0.3]
    if shift_limit is None:
        shift_limit = [-0.1, 0.1]
    if seed is None:
        return _ecvl.OpticalDistortion(src, dst, distort_limit, shift_limit,
                                       interp, border_type, border_value)
    return _ecvl.OpticalDistortion(src, dst, distort_limit, shift_limit,
                                   interp, border_type, border_value, seed)


def Salt(src, dst, p, per_channel=False, seed=None):
    """\
    Add salt noise (white pixels) to the input image.

    :param src: input image
    :param dst: output image
    :param p: probability of replacing a pixel with salt noise
    :param per_channel: if True, apply channel-wide noise
    :param seed: seed for the random number generator
    :return: None
    """
    if seed is None:
        return _ecvl.Salt(src, dst, p, per_channel)
    return _ecvl.Salt(src, dst, p, per_channel, seed)


def Pepper(src, dst, p, per_channel=False, seed=None):
    """\
    Add pepper noise (black pixels) to the input image.

    :param src: input image
    :param dst: output image
    :param p: probability of replacing a pixel with pepper noise
    :param per_channel: if True, apply channel-wide noise
    :param seed: seed for the random number generator
    :return: None
    """
    if seed is None:
        return _ecvl.Pepper(src, dst, p, per_channel)
    return _ecvl.Pepper(src, dst, p, per_channel, seed)


def SaltAndPepper(src, dst, p, per_channel=False, seed=None):
    """\
    Add salt and pepper noise (white and black pixels) to the input image.

    White and black pixels are equally likely.

    :param src: input image
    :param dst: output image
    :param p: probability of replacing a pixel with salt or pepper noise
    :param per_channel: if True, apply channel-wide noise
    :param seed: seed for the random number generator
    :return: None
    """
    if seed is None:
        return _ecvl.SaltAndPepper(src, dst, p, per_channel)
    return _ecvl.SaltAndPepper(src, dst, p, per_channel, seed)


def SliceTimingCorrection(src, dst, odd=False, down=False):
    """\
    Correct each voxel's time-series.

    Slice timing correction works by using (Hanning-windowed) sinc
    interpolation to shift each time-series by an appropriate fraction of a TR
    relative to the middle of the TR period. The default slice order
    acquisition is from the bottom of the brain to the top.

    :param src: input image. Channels must be ``"xyzt"`` and the image must
      have spacings (distance between consecutive voxels on each dimension)
    :param dst: output image
    :param odd: True if odd slices were acquired with interleaved order
      ``(0, 2, 4, ..., 1, 3, 5, ...)``
    :param down: True if down slices were acquired from the top of the brain
      to the bottom
    """
    return _ecvl.SliceTimingCorrection(src, dst, odd, down)


def Moments(src, dst, order=3, type_=DataType.float64):
    """\
    Calculate raw image moments of the source image up to the specified order.

    Moments are stored in the output image in the same order as for source
    channels. The output image will be on the same device as the source image.

    :param src: input image. It must be a grayscale (ColorType.GRAY) or a
      data (ColorType.none) image.
    :param dst: output image (ColorType.none) containing the computed raw
      image moments. The size of the Image will be (order + 1, order + 1)
    :param order: moments order
    :param type_: data type for the output image

    """
    return _ecvl.Moments(src, dst, order, type_)


def CentralMoments(src, moments, center, order=3, type_=DataType.float64):
    """\
    Calculate central moments of the source image up to the specified order.

    :param src: input image. It must be a grayscale (ColorType.GRAY) or a
      data (ColorType.none) image.
    :param moments: output data (ColorType.none) image containing the computed
      moments. The size of the Image will be (order + 1, order + 1)
    :param center: center coordinates (list of floats) ``len(center)`` and
      ``src.dims_`` must match. The source axes order must be the same used to
      specify the center coordinates.
    :param order: moments order
    :param type_: data type for the output image
    """
    return _ecvl.CentralMoments(src, moments, center, order, type_)


def DrawEllipse(src, center, axes, angle, color, thickness=1):
    """\
    Draw an ellipse over the specified Image.

    :param src: input (and output) image.
    :param center: center of the ellipse
    :param axes: ellipse axes half size
    :param angle: rotation angle of the ellipse
    :param color: ellipse color, e.g., ``[255]``, ``[5, 5, 5]`` (RGB)
    :param thickness: ellipse border thickness. If negative all the pixels of
      the ellipse will be filled with the specified color value.
    """
    return _ecvl.DrawEllipse(src, center, axes, angle, color, thickness)


def DropColorChannel(src):
    """\
    Remove color channel from the input image.

    Remove the color channel ("c") from the specified input image, modifying
    all other attributes accordingly. This function can only be applied to
    images with ``ColorType.GRAY``, i.e., having the color channel dimension
    equal to 1.

    :param src: input image.
    """
    return _ecvl.DropColorChannel(src)


def Normalize(src, dst, mean, std):
    """\
    Normalize the input image with mean and standard deviation.

    For each pixel, subtract mean and divide by std. Useful to normalize a
    dataset, getting the data within a range.

    For xyc input images, ``mean`` and ``std`` can be lists of floats,
    representing separate mean and standard deviation for each color channel.

    :param src: input image.
    :param dst: output image.
    :param mean: mean to use for normalization.
    :param std: standard deviation to use for normalization.
    """
    return _ecvl.Normalize(src, dst, mean, std)


def CenterCrop(src, dst, size):
    """\
    Crop the given image at the center.

    :param src: input image.
    :param dst: output image.
    :param size: list of integers [w, h] specifying the desired output size
    """
    return _ecvl.CenterCrop(src, dst, size)


def ScaleTo(src, dst, new_min, new_max):
    """\
    Linearly scale an Image to a new range.

    :param src: input image.
    :param dst: output image.
    :param new_min: new minimum value
    :param new_max: new maximum value
    """
    return _ecvl.ScaleTo(src, dst, new_min, new_max)



# Copyright (c) 2019-2021 CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""\
Reads a DeepHealth dataset.
"""

import argparse
import sys

import pyecvl.ecvl as ecvl


def main(args):
    print("reading", args.in_fn)
    d = ecvl.Dataset(args.in_fn)
    print("name:", d.name_)
    print("description:", d.description_)
    print("classes:", d.classes_)
    print("features:", d.features_)
    print("n. samples:", len(d.samples_))
    print("  training:", len(d.split_.training_))
    print("  validation:", len(d.split_.validation_))
    print("  test:", len(d.split_.test_))
    print("loading first sample image")
    print("  location:", d.samples_[0].location_)
    print("  label_path:", d.samples_[0].label_path_)
    img = d.samples_[0].LoadImage()
    out_path = "img0.png"
    print("saving first sample image as", out_path)
    ecvl.ImWrite(out_path, img)

    # check setter for completeness, but why would anyone want to do this?
    loc = ["/foo/bar"]
    d.samples_[0].location_ = loc
    assert d.samples_[0].location_ == loc
    lp = "/foo/tar"
    d.samples_[0].label_path_ = lp
    assert d.samples_[0].label_path_ == lp


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("in_fn", metavar="INPUT_PATH")
    main(parser.parse_args(sys.argv[1:]))
